"""
This test application exists to confirm how Zappa handles WSGI application
_responses_ when Binary Support is enabled.
"""

import gzip
import json

from flask import Flask, Response

app = Flask(__name__)


@app.route("/textplain_mimetype_response1", methods=["GET"])
def text_mimetype_response_1():
    return Response(response="OK", mimetype="text/plain")


@app.route("/textarbitrary_mimetype_response1", methods=["GET"])
def text_mimetype_response_2():
    return Response(response="OK", mimetype="text/arbitary")


@app.route("/json_mimetype_response1", methods=["GET"])
def json_mimetype_response_1():
    return Response(response=json.dumps({"some": "data"}), mimetype="application/json")


@app.route("/arbitrarybinary_mimetype_response1", methods=["GET"])
def arbitrary_mimetype_response_1():
    return Response(response=b"some binary data", mimetype="arbitrary/binary_mimetype")


@app.route("/arbitrarybinary_mimetype_response2", methods=["GET"])
def arbitrary_mimetype_response_3():
    return Response(response="doesnt_matter", mimetype="definitely_not_text")


@app.route("/content_encoding_header_json1", methods=["GET"])
def response_with_content_encoding_1():
    return Response(
        response=gzip.compress(json.dumps({"some": "data"}).encode()),
        mimetype="application/json",
        headers={"Content-Encoding": "gzip"},
    )


@app.route("/content_encoding_header_textarbitrary1", methods=["GET"])
def response_with_content_encoding_2():
    return Response(
        response=b"OK",
        mimetype="text/arbitrary",
        headers={"Content-Encoding": "something_arbitrarily_binary"},
    )


@app.route("/content_encoding_header_textarbitrary2", methods=["GET"])
def response_with_content_encoding_3():
    return Response(
        response="OK",
        mimetype="text/arbitrary",
        headers={"Content-Encoding": "with_content_type_but_not_bytes_response"},
    )


@app.route("/userdefined_additional_mimetype_response1", methods=["GET"])
def response_with_userdefined_addtional_mimetype():
    return Response(
        response="OK",
        mimetype="application/vnd.oai.openapi",
    )



API_STAGE = "dev"
APP_FUNCTION = "app"
APP_MODULE = "tests.test_wsgi_root_log_level_app"
BINARY_SUPPORT = False
CONTEXT_HEADER_MAPPINGS = {}
DEBUG = "True"
DJANGO_SETTINGS = None
DOMAIN = "api.example.com"
ENVIRONMENT_VARIABLES = {}
LOG_LEVEL = "INFO"
PROJECT_NAME = "wsgi_root_log_level"
COGNITO_TRIGGER_MAPPING = {}



from zappa.asynchronous import task

try:
    from urllib.parse import parse_qs
except ImportError:
    from cgi import parse_qs

try:
    from html import escape
except ImportError:
    from cgi import escape


def hello_world(environ, start_response):
    parameters = parse_qs(environ.get("QUERY_STRING", ""))
    if "subject" in parameters:
        subject = escape(parameters["subject"][0])
    else:
        subject = "World"
    start_response("200 OK", [("Content-Type", "text/html")])
    return [
        """Hello {subject!s}
    Hello {subject!s}!

""".format(
            **{"subject": subject}
        )
    ]


def schedule_me():
    return "Hello!"


@task
def async_me(arg1, **kwargs):
    return "run async when on lambda %s%s" % (arg1, kwargs.get("foo", ""))


@task(remote_aws_lambda_function_name="test-app-dev", remote_aws_region="us-east-1")
def remote_async_me(arg1, **kwargs):
    return "run async always on lambda %s%s" % (arg1, kwargs.get("foo", ""))


def callback(self):
    print("this is a callback")


def prebuild_me():
    print("this is a prebuild script")



# -*- coding: utf8 -*-
import os
import random
import string
import unittest
from io import BytesIO

import mock

from zappa.cli import ZappaCLI
from zappa.core import Zappa
from zappa.handler import LambdaHandler
from zappa.utilities import add_event_source, remove_event_source

from .utils import placebo_session


def random_string(length):
    return "".join(random.choice(string.printable) for _ in range(length))


class TestZappa(unittest.TestCase):
    def setUp(self):
        self.sleep_patch = mock.patch("time.sleep", return_value=None)
        # Tests expect us-east-1.
        # If the user has set a different region in env variables, we set it aside for now and use us-east-1
        self.users_current_region_name = os.environ.get("AWS_DEFAULT_REGION", None)
        os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
        if not os.environ.get("PLACEBO_MODE") == "record":
            self.sleep_patch.start()

    def tearDown(self):
        if not os.environ.get("PLACEBO_MODE") == "record":
            self.sleep_patch.stop()
        del os.environ["AWS_DEFAULT_REGION"]
        if self.users_current_region_name is not None:
            # Give the user their AWS region back, we're done testing with us-east-1.
            os.environ["AWS_DEFAULT_REGION"] = self.users_current_region_name

    @placebo_session
    def test_upload_remove_s3(self, session):
        bucket_name = "test_zappa_upload_s3"
        z = Zappa(session)
        zip_path = z.create_lambda_zip(minify=False)
        res = z.upload_to_s3(zip_path, bucket_name)
        self.assertTrue(res)
        s3 = session.resource("s3")

        # will throw ClientError with 404 if bucket doesn't exist
        s3.meta.client.head_bucket(Bucket=bucket_name)

        # will throw ClientError with 404 if object doesn't exist
        s3.meta.client.head_object(
            Bucket=bucket_name,
            Key=zip_path,
        )
        res = z.remove_from_s3(zip_path, bucket_name)
        self.assertTrue(res)

        fail = z.upload_to_s3("/tmp/this_isnt_real", bucket_name)
        self.assertFalse(fail)

        # Will graciouly handle quirky S3 behavior on 'us-east-1' region name'
        z.aws_region = "us-east-1"
        res = z.upload_to_s3(zip_path, bucket_name)
        os.remove(zip_path)
        self.assertTrue(res)

    @placebo_session
    def test_copy_on_s3(self, session):
        bucket_name = "test_zappa_upload_s3"
        z = Zappa(session)
        zip_path = z.create_lambda_zip(minify=False)
        res = z.upload_to_s3(zip_path, bucket_name)
        self.assertTrue(res)
        s3 = session.resource("s3")

        # will throw ClientError with 404 if bucket doesn't exist
        s3.meta.client.head_bucket(Bucket=bucket_name)

        # will throw ClientError with 404 if object doesn't exist
        s3.meta.client.head_object(
            Bucket=bucket_name,
            Key=zip_path,
        )
        zp = "copy_" + zip_path
        res = z.copy_on_s3(zip_path, zp, bucket_name)
        os.remove(zip_path)
        self.assertTrue(res)

    @placebo_session
    def test_create_lambda_function_s3(self, session):
        bucket_name = "lmbda"
        zip_path = "Spheres-dev-1454694878.zip"

        z = Zappa(session)
        z.aws_region = "us-east-1"
        z.load_credentials(session)
        z.credentials_arn = "arn:aws:iam::12345:role/ZappaLambdaExecution"

        arn = z.create_lambda_function(
            bucket=bucket_name,
            s3_key=zip_path,
            function_name="test_lmbda_function55",
            handler="runme.lambda_handler",
        )

        arn = z.update_lambda_function(
            bucket=bucket_name,
            s3_key=zip_path,
            function_name="test_lmbda_function55",
        )

    @placebo_session
    def test_create_lambda_function_local(self, session):
        bucket_name = "lmbda"
        local_file = "Spheres-dev-1454694878.zip"

        z = Zappa(session)
        z.aws_region = "us-east-1"
        z.load_credentials(session)
        z.credentials_arn = "arn:aws:iam::12345:role/ZappaLambdaExecution"

        arn = z.create_lambda_function(
            bucket=bucket_name,
            local_zip=local_file,
            function_name="test_lmbda_function55",
            handler="runme.lambda_handler",
        )

        arn = z.update_lambda_function(
            bucket=bucket_name,
            local_zip=local_file,
            function_name="test_lmbda_function55",
        )

    @placebo_session
    def test_create_lambda_function_docker(self, session):
        bucket_name = "lmbda"
        docker_image_uri = "docker_image_uri"

        z = Zappa(session)
        z.aws_region = "us-east-1"
        z.load_credentials(session)
        z.credentials_arn = "arn:aws:iam::12345:role/ZappaLambdaExecution"

        arn = z.create_lambda_function(
            bucket=bucket_name,
            docker_image_uri=docker_image_uri,
            function_name="test_lmbda_function55",
        )

        arn = z.update_lambda_function(
            bucket=bucket_name,
            docker_image_uri=docker_image_uri,
            function_name="test_lmbda_function55",
        )

    @placebo_session
    def test_rollback_lambda_function_version(self, session):
        z = Zappa(session)
        z.credentials_arn = "arn:aws:iam::724336686645:role/ZappaLambdaExecution"

        function_name = "django-helloworld-unicode"
        too_many_versions = z.rollback_lambda_function_version(function_name, 99999)
        self.assertFalse(too_many_versions)

        function_arn = z.rollback_lambda_function_version(function_name, 1)

    @placebo_session
    def test_rollback_lambda_function_version_docker(self, session):
        z = Zappa(session)
        z.credentials_arn = "arn:aws:iam::724336686645:role/ZappaLambdaExecution"

        function_name = "django-helloworld-unicode"

        with self.assertRaises(NotImplementedError):
            z.rollback_lambda_function_version(function_name)

    @placebo_session
    def test_invoke_lambda_function(self, session):
        z = Zappa(session)
        z.credentials_arn = "arn:aws:iam::724336686645:role/ZappaLambdaExecution"

        function_name = "django-helloworld-unicode"
        payload = '{"event": "hello"}'
        response = z.invoke_lambda_function(function_name, payload)

    @placebo_session
    def test_create_iam_roles(self, session):
        z = Zappa(session)
        arn, updated = z.create_iam_roles()
        self.assertEqual(arn, "arn:aws:iam::123:role/{}".format(z.role_name))

    @placebo_session
    def test_get_api_url(self, session):
        z = Zappa(session)
        z.credentials_arn = "arn:aws:iam::724336686645:role/ZappaLambdaExecution"
        url = z.get_api_url("Spheres-demonstration", "demonstration")

    @placebo_session
    def test_fetch_logs(self, session):
        z = Zappa(session)
        z.credentials_arn = "arn:aws:iam::12345:role/ZappaLambdaExecution"
        events = z.fetch_logs("Spheres-demonstration")
        self.assertTrue(events is not None)

    ##
    # Handler
    ##

    @placebo_session
    def test_handler(self, session):
        # Init will test load_remote_settings
        lh = LambdaHandler("test_settings", session=session)

        # Annoyingly, this will fail during record, but
        # the result will actually be okay to use in playback.
        # See: https://github.com/garnaat/placebo/issues/48
        self.assertEqual(os.environ["hello"], "world")

        event = {
            "body": None,
            "headers": {},
            "params": {
                "parameter_1": "asdf1",
                "parameter_2": "asdf2",
            },
            "method": "GET",
            "query": {},
        }
        lh.handler(event, None)

        # Test scheduled event
        event = {
            "account": "72333333333",
            "region": "us-east-1",
            "detail": {},
            "detail-type": "Scheduled Event",
            "source": "aws.events",
            "version": "0",
            "time": "2016-05-10T21:05:39Z",
            "id": "0d6a6db0-d5e7-4755-93a0-750a8bf49d55",
            "resources": ["arn:aws:events:us-east-1:72333333333:rule/tests.test_app.schedule_me"],
        }
        lh.handler(event, None)

        # Test command event
        event = {
            "account": "72333333333",
            "region": "us-east-1",
            "detail": {},
            "command": "test_settings.command",
            "source": "aws.events",
            "version": "0",
            "time": "2016-05-10T21:05:39Z",
            "id": "0d6a6db0-d5e7-4755-93a0-750a8bf49d55",
            "resources": ["arn:aws:events:us-east-1:72333333333:rule/tests.test_app.schedule_me"],
        }
        lh.handler(event, None)

        # Test command for async event
        event = {
            "account": "72333333333",
            "region": "us-east-1",
            "detail": {},
            "command": "zappa.asynchronous.route_lambda_task",
            "task_path": "tests.test_app.async_me",
            "args": ["xxx"],
            "kwargs": {},
            "source": "aws.events",
            "version": "0",
            "time": "2016-05-10T21:05:39Z",
            "id": "0d6a6db0-d5e7-4755-93a0-750a8bf49d55",
        }
        self.assertEqual("run async when on lambda xxx", lh.handler(event, None))
        event["kwargs"] = {"foo": "bar"}
        self.assertEqual("run async when on lambda xxxbar", lh.handler(event, None))

        # Test raw_command event
        event = {
            "account": "72333333333",
            "region": "us-east-1",
            "detail": {},
            "raw_command": 'print("check one two")',
            "source": "aws.events",
            "version": "0",
            "time": "2016-05-10T21:05:39Z",
            "id": "0d6a6db0-d5e7-4755-93a0-750a8bf49d55",
            "resources": ["arn:aws:events:us-east-1:72333333333:rule/tests.test_app.schedule_me"],
        }
        lh.handler(event, None)

        # Test AWS S3 event
        event = {
            "account": "72333333333",
            "region": "us-east-1",
            "detail": {},
            "Records": [{"s3": {"configurationId": "test_project:test_settings.aws_s3_event"}}],
            "source": "aws.events",
            "version": "0",
            "time": "2016-05-10T21:05:39Z",
            "id": "0d6a6db0-d5e7-4755-93a0-750a8bf49d55",
            "resources": ["arn:aws:events:us-east-1:72333333333:rule/tests.test_app.schedule_me"],
        }
        self.assertEqual("AWS S3 EVENT", lh.handler(event, None))

        # Test AWS SNS event
        event = {
            "account": "72333333333",
            "region": "us-east-1",
            "detail": {},
            "Records": [
                {
                    "EventVersion": "1.0",
                    "EventSource": "aws:sns",
                    "EventSubscriptionArn": "arn:aws:sns:EXAMPLE",
                    "Sns": {
                        "SignatureVersion": "1",
                        "Timestamp": "1970-01-01T00:00:00.000Z",
                        "Signature": "EXAMPLE",
                        "SigningCertUrl": "EXAMPLE",
                        "MessageId": "95df01b4-ee98-5cb9-9903-4c221d41eb5e",
                        "Message": "Hello from SNS!",
                        "Subject": "TestInvoke",
                        "Type": "Notification",
                        "UnsubscribeUrl": "EXAMPLE",
                        "TopicArn": "arn:aws:sns:1",
                        "MessageAttributes": {
                            "Test": {"Type": "String", "Value": "TestString"},
                            "TestBinary": {"Type": "Binary", "Value": "TestBinary"},
                        },
                    },
                }
            ],
        }
        self.assertEqual("AWS SNS EVENT", lh.handler(event, None))

        # Test AWS SNS event
        event = {
            "account": "72333333333",
            "region": "us-east-1",
            "detail": {},
            "Records": [
                {
                    "EventVersion": "1.0",
                    "EventSource": "aws:sns",
                    "EventSubscriptionArn": "arn:aws:sns:EXAMPLE",
                    "Sns": {
                        "SignatureVersion": "1",
                        "Timestamp": "1970-01-01T00:00:00.000Z",
                        "Signature": "EXAMPLE",
                        "SigningCertUrl": "EXAMPLE",
                        "MessageId": "95df01b4-ee98-5cb9-9903-4c221d41eb5e",
                        "Message": '{"args": ["arg1", "arg2"], "command": "zappa.asynchronous.route_sns_task", '
                        '"task_path": "test_settings.aws_async_sns_event", "kwargs": {"arg3": "varg3"}}',
                        "Subject": "TestInvoke",
                        "Type": "Notification",
                        "UnsubscribeUrl": "EXAMPLE",
                        "MessageAttributes": {
                            "Test": {"Type": "String", "Value": "TestString"},
                            "TestBinary": {"Type": "Binary", "Value": "TestBinary"},
                        },
                    },
                }
            ],
        }
        self.assertEqual("AWS ASYNC SNS EVENT", lh.handler(event, None))

        # Test AWS DynamoDB event
        event = {
            "Records": [
                {
                    "eventID": "1",
                    "eventVersion": "1.0",
                    "dynamodb": {
                        "Keys": {"Id": {"N": "101"}},
                        "NewImage": {"Message": {"S": "New item!"}, "Id": {"N": "101"}},
                        "StreamViewType": "NEW_AND_OLD_IMAGES",
                        "SequenceNumber": "111",
                        "SizeBytes": 26,
                    },
                    "awsRegion": "us-west-2",
                    "eventName": "INSERT",
                    "eventSourceARN": "arn:aws:dynamodb:1",
                    "eventSource": "aws:dynamodb",
                }
            ]
        }
        self.assertEqual("AWS DYNAMODB EVENT", lh.handler(event, None))

        # Test AWS kinesis event
        event = {
            "Records": [
                {
                    "eventID": "shardId-000000000000:49545115243490985018280067714973144582180062593244200961",
                    "eventVersion": "1.0",
                    "kinesis": {
                        "partitionKey": "partitionKey-3",
                        "data": "SGVsbG8sIHRoaXMgaXMgYSB0ZXN0IDEyMy4=",
                        "kinesisSchemaVersion": "1.0",
                        "sequenceNumber": "49545115243490985018280067714973144582180062593244200961",
                    },
                    "invokeIdentityArn": "arn:aws:iam::EXAMPLE",
                    "eventName": "aws:kinesis:record",
                    "eventSourceARN": "arn:aws:kinesis:1",
                    "eventSource": "aws:kinesis",
                    "awsRegion": "us-east-1",
                }
            ]
        }
        self.assertEqual("AWS KINESIS EVENT", lh.handler(event, None))

        # Test AWS SQS event
        event = {
            "Records": [
                {
                    "messageId": "c80e8021-a70a-42c7-a470-796e1186f753",
                    "receiptHandle": "AQEBJQ+/u6NsnT5t8Q/VbVxgdUl4TMKZ5FqhksRdIQvLBhwNvADoBxYSOVeCBXdnS9P+erlTtwEALHsnBXynkfPLH3BOUqmgzP25U8kl8eHzq6RAlzrSOfTO8ox9dcp6GLmW33YjO3zkq5VRYyQlJgLCiAZUpY2D4UQcE5D1Vm8RoKfbE+xtVaOctYeINjaQJ1u3mWx9T7tork3uAlOe1uyFjCWU5aPX/1OHhWCGi2EPPZj6vchNqDOJC/Y2k1gkivqCjz1CZl6FlZ7UVPOx3AMoszPuOYZ+Nuqpx2uCE2MHTtMHD8PVjlsWirt56oUr6JPp9aRGo6bitPIOmi4dX0FmuMKD6u/JnuZCp+AXtJVTmSHS8IXt/twsKU7A+fiMK01NtD5msNgVPoe9JbFtlGwvTQ==",
                    "body": BytesIO('{"foo":"bar"}'.encode()),
                    "attributes": {
                        "ApproximateReceiveCount": "3",
                        "SentTimestamp": "1529104986221",
                        "SenderId": "594035263019",
                        "ApproximateFirstReceiveTimestamp": "1529104986230",
                    },
                    "messageAttributes": {},
                    "md5OfBody": "9bb58f26192e4ba00f01e2e7b136bbd8",
                    "eventSource": "aws:sqs",
                    "eventSourceARN": "arn:aws:sqs:1",
                    "awsRegion": "us-east-1",
                }
            ]
        }
        self.assertEqual("AWS SQS EVENT", lh.handler(event, None))

        # Test Authorizer event
        event = {
            "authorizationToken": "hubtoken1",
            "methodArn": "arn:aws:execute-api:us-west-2:1234:xxxxx/dev/GET/v1/endpoint/param",
            "type": "TOKEN",
        }
        self.assertEqual("AUTHORIZER_EVENT", lh.handler(event, None))

        # Ensure Zappa does return 401 if no function was defined.
        lh.settings.AUTHORIZER_FUNCTION = None
        with self.assertRaisesRegex(Exception, "Unauthorized"):
            lh.handler(event, None)

        # Unhandled event
        event = {
            "Records": [
                {
                    "eventID": "shardId-000000000000:49545115243490985018280067714973144582180062593244200961",
                    "eventVersion": "1.0",
                    "kinesis": {
                        "partitionKey": "partitionKey-3",
                        "data": "SGVsbG8sIHRoaXMgaXMgYSB0ZXN0IDEyMy4=",
                        "kinesisSchemaVersion": "1.0",
                        "sequenceNumber": "49545115243490985018280067714973144582180062593244200961",
                    },
                    "eventSourceARN": "bad:arn:1",
                }
            ]
        }
        self.assertIsNone(lh.handler(event, None))

    ##
    # CLI
    ##

    @placebo_session
    def test_cli_aws(self, session):
        zappa_cli = ZappaCLI()
        zappa_cli.api_stage = "ttt888"
        zappa_cli.api_key_required = True
        zappa_cli.authorization_type = "NONE"
        zappa_cli.load_settings("test_settings.json", session)
        zappa_cli.zappa.credentials_arn = "arn:aws:iam::12345:role/ZappaLambdaExecution"
        zappa_cli.deploy()
        zappa_cli.update()
        zappa_cli.rollback(1)
        zappa_cli.tail(since=0, filter_pattern="", keep_open=False)
        zappa_cli.schedule()
        zappa_cli.unschedule()
        zappa_cli.undeploy(no_confirm=True, remove_logs=True)

    @placebo_session
    def test_cli_aws_status(self, session):
        zappa_cli = ZappaCLI()
        zappa_cli.api_stage = "ttt888"
        zappa_cli.load_settings("test_settings.json", session)
        zappa_cli.api_stage = "devor"
        zappa_cli.lambda_name = "baby-flask-devor"
        zappa_cli.zappa.credentials_arn = "arn:aws:iam::12345:role/ZappaLambdaExecution"
        resp = zappa_cli.status()

    ##
    # Let's Encrypt / ACME
    ##

    ##
    # Django
    ##

    ##
    # Util / Misc
    ##

    @placebo_session
    def test_add_event_source(self, session):
        event_source = {"arn": "blah:blah:blah:blah", "events": ["s3:ObjectCreated:*"]}
        # Sanity. This should fail.
        try:
            es = add_event_source(event_source, "blah:blah:blah:blah", "test_settings.callback", session)
            self.fail("Success should have failed.")
        except ValueError:
            pass

        event_source = {"arn": "s3:s3:s3:s3", "events": ["s3:ObjectCreated:*"]}
        add_event_source(
            event_source,
            "lambda:lambda:lambda:lambda",
            "test_settings.callback",
            session,
            dry=True,
        )
        remove_event_source(
            event_source,
            "lambda:lambda:lambda:lambda",
            "test_settings.callback",
            session,
            dry=True,
        )
        event_source = {
            "arn": "sqs:sqs:sqs:sqs",
            "events": [
                {
                    "function": "function_name",
                    "event_source": {
                        "arn": "dummy",
                        "batch_size": 100,
                        "batch_window": 10,
                        "functionResponseType": "ReportBatchItemFailures",
                        "enabled": True,
                    },
                }
            ],
        }
        add_event_source(
            event_source,
            "sqs:sqs:sqs:sqs",
            "test_settings.callback",
            session,
            dry=True,
        )
        remove_event_source(
            event_source,
            "sqs:sqs:sqs:sqs",
            "test_settings.callback",
            session,
            dry=True,
        )
        event_source = {
            "arn": "sqs:sqs:sqs:sqs",
            "events": [
                {
                    "function": "function_name",
                    "event_source": {
                        "arn": "dummy",
                        "batch_size": 100,
                        "functionResponseType": "ReportBatchItemFailures",
                        "enabled": True,
                    },
                }
            ],
        }
        add_event_source(
            event_source,
            "sqs:sqs:sqs:sqs",
            "test_settings.callback",
            session,
            dry=True,
        )
        remove_event_source(
            event_source,
            "sqs:sqs:sqs:sqs",
            "test_settings.callback",
            session,
            dry=True,
        )
        # get_event_source_status(event_source, 'lambda:lambda:lambda:lambda', 'test_settings.callback', session, dry=True)

    @placebo_session
    def test_cognito_trigger(self, session):
        z = Zappa(session)
        z.update_cognito(
            "Zappa-Trigger-Test",
            "us-east-1_9jUv74DH8",
            {"PreSignUp": "test.tasks.pre_signup"},
            "arn:aws:lambda:us-east-1:12345:function:Zappa-Trigger-Test",
        )

    @placebo_session
    def test_cognito_trigger_existing(self, session):
        z = Zappa(session)
        z.update_cognito(
            "Zappa-Trigger-Test",
            "us-east-1_9jUv74DH8",
            {"PreSignUp": "test.tasks.pre_signup"},
            "arn:aws:lambda:us-east-1:12345:function:Zappa-Trigger-Test",
        )

    @placebo_session
    def test_cli_cognito_triggers(self, session):
        zappa_cli = ZappaCLI()
        zappa_cli.api_stage = "ttt888"
        zappa_cli.api_key_required = True
        zappa_cli.load_settings("test_settings.json", session)
        zappa_cli.lambda_arn = "arn:aws:lambda:us-east-1:12345:function:Zappa-Trigger-Test"
        zappa_cli.update_cognito_triggers()

    @placebo_session
    def test_cognito_trigger_existing_UnusedAccountValidityDays(self, session):
        z = Zappa(session)
        z.update_cognito(
            "Zappa-Trigger-Test",
            "us-east-1_9jUv74DH8",
            {"PreSignUp": "test.tasks.pre_signup"},
            "arn:aws:lambda:us-east-1:12345:function:Zappa-Trigger-Test",
        )


if __name__ == "__main__":
    unittest.main()



from configparser import ConfigParser
from io import open
from pathlib import Path

from setuptools import setup

from zappa import __version__

with open("README.md", encoding="utf-8") as readme_file:
    long_description = readme_file.read()

pipfile = ConfigParser()
pipfile.read(Path(__file__).parent.resolve() / "Pipfile")
required = [
    "{}{}".format(name, version.strip('"')) if version != '"*"' else name for name, version in pipfile["packages"].items()
]

setup(
    name="zappa",
    version=__version__,
    packages=["zappa"],
    install_requires=required,
    python_requires=">=3.9",
    include_package_data=True,
    license="MIT License",
    description="Server-less Python Web Services for AWS Lambda and API Gateway",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zappa/Zappa",
    author="Rich Jones",
    author_email="rich@openwatch.net",
    entry_points={
        "console_scripts": [
            "zappa=zappa.cli:handle",
            "z=zappa.cli:handle",
        ],
    },
    classifiers=[
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Framework :: Django",
        "Framework :: Django :: 4.2",
        "Framework :: Django :: 5.2",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



import re
import unittest
from copy import copy
from os import environ, path

DIR = path.realpath(path.dirname(__file__))


class TestDocs(unittest.TestCase):
    def test_readmetoc(self):
        # this test serves as a reminder to update the README toc.
        # If you're here because the test is failing, you can update the toc by running `doctoc README.md`

        # doctoc: https://github.com/thlorenz/doctoc
        # More information here: https://github.com/Miserlou/Zappa/issues/1228
        # borrowed in part from https://github.com/PrzemekWirkus/git-toc/blob/729fe22417d2b310b4131bb592cd18b235633a8c/gittoc/gittoc.py (Apache license)

        start_marker = "<!-- START doctoc generated TOC please keep comment here to allow auto update -->\n"
        end_marker = "<!-- END doctoc generated TOC please keep comment here to allow auto update -->\n"
        template = """{prologue}{start_marker}<!-- DON'T EDIT THIS SECTION, INSTEAD RE-RUN doctoc TO UPDATE -->


{toc}

{end_marker}{epilogue}"""

        prologue = ""
        toc = []
        epilogue = ""

        state = "prologue"

        with open(path.join(path.dirname(DIR), "README.md")) as f:
            contents = f.readlines()
            old_readme = copy(contents)
            state = "prologue"

            for line in contents:
                if state == "prologue":
                    if line == start_marker:
                        state = "toc"
                    else:
                        prologue = prologue + line

                elif state == "toc":
                    # we don't need to capture the old TOC
                    if line == end_marker:
                        state = "epilogue"

                elif state == "epilogue":
                    epilogue = epilogue + line

                    # we only capture TOC contents *after* the TOC markers
                    m = re.search("^([#]{1,6}) (.*)$", line)
                    if m is not None and m.groups():
                        header = m.group(1)
                        header_text = m.group(2)
                        header_text_strip = re.sub("[^a-zA-Z0-9-_ ]", "", header_text)

                        if header_text_strip == "":
                            # skip empty header
                            continue

                        header_text_no_spaces = header_text_strip.replace(" ", "-").lower()
                        toc_line = "  " * (len(header) - 2) + "- [%s](#%s)" % (
                            header_text,
                            header_text_no_spaces.lower(),
                        )
                        toc.append(toc_line)

        new_readme = template.format(
            toc="\n".join(toc),
            start_marker=start_marker,
            end_marker=end_marker,
            prologue=prologue,
            epilogue=epilogue,
        )

        if environ.get("ZAPPA_TEST_SAVE_README_NEW"):
            readme_test_output_filepath = path.join(path.dirname(DIR), "README.test.md")
            with open(readme_test_output_filepath, "w") as f:
                f.write(new_readme)

            msg = f"{readme_test_output_filepath} written so you can manually compare."

        else:
            msg = "You can set environ[ZAPPA_TEST_SAVE_README_NEW]=1 to generate\n" "  README.test.md to manually compare."

        self.assertEqual(
            "".join(old_readme),
            new_readme,
            "README doesn't match after regenerating TOC\n\n" "You need to run doctoc after a heading change.\n{}".format(msg),
        )



import os
import unittest

import boto3
import mock

try:
    from mock import patch
except ImportError:
    from unittest.mock import patch

from zappa.asynchronous import (
    AsyncException,
    LambdaAsyncResponse,
    SnsAsyncResponse,
    get_func_task_path,
    import_and_get_task,
)
from zappa.utilities import UnserializableJsonError


class TestZappa(unittest.TestCase):
    def setUp(self):
        self.sleep_patch = mock.patch("time.sleep", return_value=None)
        # Tests expect us-east-1.
        # If the user has set a different region in env variables, we set it aside for now and use us-east-1
        self.users_current_region_name = os.environ.get("AWS_DEFAULT_REGION", None)
        os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
        if not os.environ.get("PLACEBO_MODE") == "record":
            self.sleep_patch.start()

    def tearDown(self):
        if not os.environ.get("PLACEBO_MODE") == "record":
            self.sleep_patch.stop()
        del os.environ["AWS_DEFAULT_REGION"]
        if self.users_current_region_name is not None:
            # Give the user their AWS region back, we're done testing with us-east-1.
            os.environ["AWS_DEFAULT_REGION"] = self.users_current_region_name

    ##
    # Sanity Tests
    ##

    def test_test(self):
        self.assertTrue(True)
        self.assertFalse(False)

    def test_nofails_classes(self):
        boto_session = boto3.Session(region_name=os.environ["AWS_DEFAULT_REGION"])

        a = AsyncException()
        l = LambdaAsyncResponse(boto_session=boto_session)
        # s = SnsAsyncResponse()
        s = SnsAsyncResponse(arn="arn:abc:def", boto_session=boto_session)

    def test_nofails_funcs(self):
        funk = import_and_get_task("tests.test_app.async_me")
        get_func_task_path(funk)
        self.assertEqual(funk.__name__, "async_me")

    ##
    # Functional tests
    ##
    def test_sync_call(self):
        funk = import_and_get_task("tests.test_app.async_me")
        self.assertEqual(funk.sync("123"), "run async when on lambda 123")

    def test_async_call_with_defaults(self):
        """Change a task's asynchronousity at runtime."""
        # Import the task first to make sure it is decorated whilst the
        # environment is unpatched.
        async_me = import_and_get_task("tests.test_app.async_me")
        lambda_async_mock = mock.Mock()
        lambda_async_mock.return_value.send.return_value = "Running async!"
        with mock.patch.dict("zappa.asynchronous.ASYNC_CLASSES", {"lambda": lambda_async_mock}):
            # First check that it still runs synchronously by default
            self.assertEqual(async_me("123"), "run async when on lambda 123")

            # Now patch the environment to make it look like we are running on
            # AWS Lambda
            options = {
                "AWS_LAMBDA_FUNCTION_NAME": "MyLambda",
                "AWS_REGION": "us-east-1",
            }
            with mock.patch.dict(os.environ, options):
                self.assertEqual(async_me("qux"), "Running async!")

        # And check the dispatching class got called correctly
        lambda_async_mock.assert_called_once()
        lambda_async_mock.assert_called_with(
            aws_region="us-east-1",
            capture_response=False,
            lambda_function_name="MyLambda",
        )
        lambda_async_mock.return_value.send.assert_called_with(get_func_task_path(async_me), ("qux",), {})

    def test_async_call_arg_not_json_serializable(self):
        """Exception is raised when calling an async function locally (not on aws)"""
        async_me = import_and_get_task("tests.test_app.async_me")
        unserializable_object = object()
        with self.assertRaises(UnserializableJsonError):
            async_me(unserializable_object)



import logging

from flask import Flask, request

app = Flask(__name__)


@app.route("/root-logger", methods=["GET", "POST"])
def return_request_url():
    logging.debug("debug message")
    logging.info("info message")
    logging.warning("warning message")
    logging.error("error message")
    logging.critical("critical message")
    return ""



import json
import os
import re
import tempfile
import unittest
from pathlib import Path
from typing import Tuple
from unittest import mock

from zappa.core import Zappa
from zappa.ext.django_zappa import get_django_wsgi
from zappa.utilities import (
    ApacheNCSAFormatter,
    InvalidAwsLambdaName,
    conflicts_with_a_neighbouring_module,
    contains_python_files_or_subdirs,
    detect_django_settings,
    detect_flask_apps,
    get_venv_from_python_version,
    human_size,
    is_valid_bucket_name,
    parse_s3_url,
    string_to_timestamp,
    titlecase_keys,
    validate_name,
)


class GeneralUtilitiesTestCase(unittest.TestCase):
    def setUp(self):
        self.sleep_patch = mock.patch("time.sleep", return_value=None)
        # Tests expect us-east-1.
        # If the user has set a different region in env variables, we set it aside for now and use us-east-1
        self.users_current_region_name = os.environ.get("AWS_DEFAULT_REGION", None)
        os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
        if not os.environ.get("PLACEBO_MODE") == "record":
            self.sleep_patch.start()

    def tearDown(self):
        if not os.environ.get("PLACEBO_MODE") == "record":
            self.sleep_patch.stop()
        del os.environ["AWS_DEFAULT_REGION"]
        if self.users_current_region_name is not None:
            # Give the user their AWS region back, we're done testing with us-east-1.
            os.environ["AWS_DEFAULT_REGION"] = self.users_current_region_name

    @mock.patch("zappa.core.find_packages", return_value=["package", "package.subpackage", "package.another"])
    def test_copy_editable_packages(self, mock_find_packages):
        virtual_env = Path(os.environ.get("VIRTUAL_ENV"))
        if not virtual_env:
            return self.skipTest("test_copy_editable_packages must be run in a virtualenv")

        with tempfile.TemporaryDirectory() as temp_package_dir:
            temp_package_dir = Path(temp_package_dir)

            egg_link_filename = "test-copy-editable-packages.egg-link"
            egg_link = virtual_env / "lib" / get_venv_from_python_version() / "site-packages" / egg_link_filename
            egg_path = "/some/other/directory/package"
            egg_link_open = mock.mock_open(
                read_data=egg_path.encode("utf-8")
            )  # content of "test-copy-editable-packages.egg-link"

            with mock.patch.object(Path, "open", egg_link_open), mock.patch.object(
                Path, "unlink"
            ) as mock_unlink, mock.patch.object(Path, "glob", return_value=[Path(egg_link_filename)]) as mock_glob:
                egg_links = [egg_link]
                z = Zappa()
                with mock.patch("zappa.core.copytree") as mock_copytree:
                    # Reads the first line of each egg-link file, which contains the path to the package,
                    # copies the package contents to the temporary package path, then removes the egg-link files.

                    # - find_packages() mocked to provide the list of "packages" contained in the directory (egg_path) defined in the `egg_link` file
                    # - copytree mocked to determine which files are copied
                    # - temp_package_path.glob("*.egg-link") mocked to ignore unlink (delete) of egg-link files
                    z.copy_editable_packages(egg_links, temp_package_dir)

                    # confirm that the files/directories in the path defined in egg-link file (egg_path)
                    # are copied to the temp package directory
                    # find_packages() is called on the egg_path to find the packages
                    mock_copytree.assert_called_with(
                        Path(egg_path) / "package",
                        temp_package_dir / "package",
                        metadata=False,
                        symlinks=False,
                    )
                    self.assertEqual(mock_copytree.call_count, 1)

                    # confirm that temp_package_path.glob() is called to find egg-link files
                    self.assertEqual(mock_glob.call_count, 1)

                    # confirm that the egg-link file unlinked (deleted) after copying the package
                    self.assertEqual(mock_unlink.call_count, 1)

    def test_detect_dj(self):
        # Sanity
        settings_modules = detect_django_settings()

    def test_dj_wsgi(self):
        # Sanity
        settings_modules = detect_django_settings()

        settings = """
# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os
BASE_DIR = os.path.dirname(os.path.dirname(__file__))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.7/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'alskdfjalsdkf=0*%do-ayvy*m2k=vss*$7)j8q!@u0+d^na7mi2(^!l!d'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

TEMPLATE_DEBUG = True

ALLOWED_HOSTS = []

# Application definition

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'blah.urls'
WSGI_APPLICATION = 'hackathon_starter.wsgi.application'

# Database
# https://docs.djangoproject.com/en/1.7/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

# Internationalization
# https://docs.djangoproject.com/en/1.7/topics/i18n/

LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True
        """

        djts = open("dj_test_settings.py", "w")
        djts.write(settings)
        djts.close()

        app = get_django_wsgi("dj_test_settings")
        try:
            os.remove("dj_test_settings.py")
            os.remove("dj_test_settings.pyc")
        except Exception as e:
            pass

    ##
    # Util / Misc
    ##

    def test_human_units(self):
        human_size(1)
        human_size(9999999999999)

    def test_string_to_timestamp(self):
        boo = string_to_timestamp("asdf")
        self.assertTrue(boo == 0)

        yay = string_to_timestamp("1h")
        self.assertTrue(type(yay) == int)
        self.assertTrue(yay > 0)

        yay = string_to_timestamp("4m")
        self.assertTrue(type(yay) == int)
        self.assertTrue(yay > 0)

        yay = string_to_timestamp("1mm")
        self.assertTrue(type(yay) == int)
        self.assertTrue(yay > 0)

        yay = string_to_timestamp("1mm1w1d1h1m1s1ms1us")
        self.assertTrue(type(yay) == int)
        self.assertTrue(yay > 0)

    def test_detect_dj(self):
        # Sanity
        settings_modules = detect_django_settings()

    def test_detect_flask(self):
        # Sanity
        settings_modules = detect_flask_apps()

    def test_s3_url_parser(self):
        remote_bucket, remote_file = parse_s3_url("s3://my-project-config-files/filename.json")
        self.assertEqual(remote_bucket, "my-project-config-files")
        self.assertEqual(remote_file, "filename.json")

        remote_bucket, remote_file = parse_s3_url("s3://your-bucket/account.key")
        self.assertEqual(remote_bucket, "your-bucket")
        self.assertEqual(remote_file, "account.key")

        remote_bucket, remote_file = parse_s3_url("s3://my-config-bucket/super-secret-config.json")
        self.assertEqual(remote_bucket, "my-config-bucket")
        self.assertEqual(remote_file, "super-secret-config.json")

        remote_bucket, remote_file = parse_s3_url("s3://your-secure-bucket/account.key")
        self.assertEqual(remote_bucket, "your-secure-bucket")
        self.assertEqual(remote_file, "account.key")

        remote_bucket, remote_file = parse_s3_url("s3://your-bucket/subfolder/account.key")
        self.assertEqual(remote_bucket, "your-bucket")
        self.assertEqual(remote_file, "subfolder/account.key")

        # Sad path
        remote_bucket, remote_file = parse_s3_url("/dev/null")
        self.assertEqual(remote_bucket, "")

    def test_validate_name(self):
        fname = "tests/name_scenarios.json"
        with open(fname, "r") as f:
            scenarios = json.load(f)
        for scenario in scenarios:
            value = scenario["value"]
            is_valid = scenario["is_valid"]
            if is_valid:
                assert validate_name(value)
            else:
                with self.assertRaises(InvalidAwsLambdaName) as exc:
                    validate_name(value)

    def test_contains_python_files_or_subdirs(self):
        self.assertTrue(contains_python_files_or_subdirs("tests/data"))
        self.assertTrue(contains_python_files_or_subdirs("tests/data/test2"))
        self.assertFalse(contains_python_files_or_subdirs("tests/data/test1"))

    def test_conflicts_with_a_neighbouring_module(self):
        self.assertTrue(conflicts_with_a_neighbouring_module("tests/data/test1"))
        self.assertFalse(conflicts_with_a_neighbouring_module("tests/data/test2"))

    def test_titlecase_keys(self):
        raw = {
            "hOSt": "github.com",
            "ConnECtiOn": "keep-alive",
            "UpGRAde-InSecuRE-ReQueSts": "1",
            "uSer-AGEnT": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/65.0.3325.181 Safari/537.36",
            "cONtENt-TYPe": "text/html; charset=utf-8",
            "aCCEpT": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
            "ACcePT-encoDInG": "gzip, deflate, br",
            "AcCEpT-lAnGUagE": "en-US,en;q=0.9",
        }
        transformed = titlecase_keys(raw)
        expected = {
            "Host": "github.com",
            "Connection": "keep-alive",
            "Upgrade-Insecure-Requests": "1",
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/65.0.3325.181 Safari/537.36",
            "Content-Type": "text/html; charset=utf-8",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "en-US,en;q=0.9",
        }
        self.assertEqual(expected, transformed)

    def test_is_valid_bucket_name(self):
        # Bucket names must be at least 3 and no more than 63 characters long.
        self.assertFalse(is_valid_bucket_name("ab"))
        self.assertFalse(is_valid_bucket_name("abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefhijlmn"))
        # Bucket names must not contain uppercase characters or underscores.
        self.assertFalse(is_valid_bucket_name("aaaBaaa"))
        self.assertFalse(is_valid_bucket_name("aaa_aaa"))
        # Bucket names must start with a lowercase letter or number.
        self.assertFalse(is_valid_bucket_name(".abbbaba"))
        self.assertFalse(is_valid_bucket_name("abbaba."))
        self.assertFalse(is_valid_bucket_name("-abbaba"))
        self.assertFalse(is_valid_bucket_name("ababab-"))
        # Bucket names must be a series of one or more labels. Adjacent labels are separated by a single period (.).
        # Each label must start and end with a lowercase letter or a number.
        self.assertFalse(is_valid_bucket_name("aaa..bbbb"))
        self.assertFalse(is_valid_bucket_name("aaa.-bbb.ccc"))
        self.assertFalse(is_valid_bucket_name("aaa-.bbb.ccc"))
        # Bucket names must not be formatted as an IP address (for example, 192.168.5.4).
        self.assertFalse(is_valid_bucket_name("192.168.5.4"))
        self.assertFalse(is_valid_bucket_name("127.0.0.1"))
        self.assertFalse(is_valid_bucket_name("255.255.255.255"))

        self.assertTrue(is_valid_bucket_name("valid-formed-s3-bucket-name"))
        self.assertTrue(is_valid_bucket_name("worst.bucket.ever"))


class ApacheNCSAFormatterTestCase(unittest.TestCase):
    def setUp(self):
        self.method = "GET"
        self.datetime_regex = re.compile(
            r"\d+\/(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\/\d{4}:\d{2}:\d{2}:\d{2}\s\+\d{4}"
        )
        self.agent = "myagent"

    def _build_expected_format_string(self, status_code, addtional_environ, content_length, **kwargs) -> Tuple[dict, str]:
        referer = "me"
        logname = "-"
        user = "-"
        host = "127.0.0.1"
        server_protocol = "myprot"
        environ = {
            "REMOTE_ADDR": host,
            "HTTP_USER_AGENT": self.agent,
            "HTTP_REFERER": referer,
            "SERVER_PROTOCOL": server_protocol,
            "PATH_INFO": "/my/path/",
            "REQUEST_METHOD": self.method,
        }
        environ.update(addtional_environ)
        query_string = ""
        raw_query_string = environ.get("QUERY_STRING", None)
        if raw_query_string:
            query_string = f"?{raw_query_string}"
        request = f"{self.method} {environ['PATH_INFO']}{query_string} {server_protocol}"

        regex_log_entry = f'{host} {logname} {user} [] "{request}" {status_code} {content_length} "{referer}" "{self.agent}"'
        rt_us = kwargs.get("rt_us")
        if rt_us:
            rt_seconds = int(rt_us / 1_000_000)
            regex_log_entry = f"{regex_log_entry} {rt_seconds}/{rt_us}"
        return environ, regex_log_entry

    def test_with_response_time__true(self):
        formatter = ApacheNCSAFormatter(with_response_time=True)
        expected = "format_log_with_response_time"
        actual = formatter.__name__
        self.assertEqual(actual, expected)

        status_code = 200
        content_length = 10
        rt_us = 15
        environ, expected = self._build_expected_format_string(status_code, {}, content_length, rt_us=15)
        actual = formatter(status_code, environ, content_length, rt_us=rt_us)
        self.assertRegex(actual, self.datetime_regex)
        # extract and remove matched datetime
        result = self.datetime_regex.search(actual)
        match_start, match_end = result.span()
        replace_text = actual[match_start:match_end]
        actual = actual.replace(replace_text, "")
        self.assertEqual(actual, expected)

        agent_endstring = f'"{self.agent}"'
        self.assertFalse(actual.endswith(agent_endstring))

    def test_with_response_time__true__with_querystring(self):
        formatter = ApacheNCSAFormatter(with_response_time=True)

        status_code = 200
        content_length = 10
        rt_us = 15
        additional_environ = {"QUERY_STRING": "name=hello&data=hello"}
        environ, expected = self._build_expected_format_string(status_code, additional_environ, content_length, rt_us=15)
        actual = formatter(status_code, environ, content_length, rt_us=rt_us)
        self.assertRegex(actual, self.datetime_regex)
        # extract and remove matched datetime
        result = self.datetime_regex.search(actual)
        match_start, match_end = result.span()
        replace_text = actual[match_start:match_end]
        actual = actual.replace(replace_text, "")
        self.assertEqual(actual, expected)
        agent_endstring = f'"{self.agent}"'
        self.assertFalse(actual.endswith(agent_endstring))

    def test_with_response_time__false(self):
        formatter = ApacheNCSAFormatter(with_response_time=False)

        expected = "format_log"
        actual = formatter.__name__
        self.assertEqual(actual, expected)

        status_code = 200
        content_length = 10
        environ, expected = self._build_expected_format_string(status_code, {}, content_length)
        actual = formatter(status_code, environ, content_length)
        self.assertRegex(actual, self.datetime_regex)
        # extract and remove matched datetime
        result = self.datetime_regex.search(actual)
        match_start, match_end = result.span()
        replace_text = actual[match_start:match_end]
        actual = actual.replace(replace_text, "")
        self.assertEqual(actual, expected)

        agent_endstring = f'"{self.agent}"'
        self.assertTrue(actual.endswith(agent_endstring))

    def test_with_response_time__false__with_querystring(self):
        formatter = ApacheNCSAFormatter(with_response_time=False)

        status_code = 200
        content_length = 10
        additional_environ = {"QUERY_STRING": "name=hello&data=hello"}
        environ, expected = self._build_expected_format_string(status_code, additional_environ, content_length)
        actual = formatter(status_code, environ, content_length)
        self.assertRegex(actual, self.datetime_regex)
        # extract and remove matched datetime
        result = self.datetime_regex.search(actual)
        match_start, match_end = result.span()
        replace_text = actual[match_start:match_end]
        actual = actual.replace(replace_text, "")
        self.assertEqual(actual, expected)
        agent_endstring = f'"{self.agent}"'
        self.assertTrue(actual.endswith(agent_endstring))



import calendar
import datetime
import json
import logging
import os
import re
import shutil
import stat
import sys
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

import boto3
import botocore
import durationpy

LOG = logging.getLogger(__name__)


class UnserializableJsonError(TypeError):
    """Exception class for JSON encoding errors"""

    pass


##
# Settings / Packaging
##

# mimetypes starting with entries defined here are considered as TEXT when BINARTY_SUPPORT is True.
# - Additional TEXT mimetypes may be defined with the 'ADDITIONAL_TEXT_MIMETYPES' setting.
DEFAULT_TEXT_MIMETYPES = (
    "text/",
    "application/json",  # RFC 4627
    "application/javascript",  # RFC 4329
    "application/ecmascript",  # RFC 4329
    "application/xml",  # RFC 3023
    "application/xml-external-parsed-entity",  # RFC 3023
    "application/xml-dtd",  # RFC 3023
    "image/svg+xml",  # RFC 3023
)


def copytree(
    src: Path,
    dst: Path,
    metadata: bool = True,
    symlinks: bool = False,
    ignore: Optional[Callable[[Any, list[str]], set[str]]] = None,
) -> None:
    """
    This is a contributed re-implementation of 'copytree' that
    should work with the exact same behavior on multiple platforms.

    When `metadata` is False, file metadata such as permissions and modification
    times are not copied.
    """

    def copy_file(src_path: Path, dst_path: Path, item: str) -> None:
        s = src_path / item
        d = dst_path / item

        if symlinks and s.is_symlink():  # pragma: no cover
            if d.exists():
                d.unlink()
            d.symlink_to(s.readlink())
            if metadata:
                st = s.lstat()
                mode = stat.S_IMODE(st.st_mode)
                try:
                    os.chmod(str(d), mode)
                except Exception:
                    LOG.warning(f"Unable to perform chmod on: {d}")
        elif s.is_dir():
            copytree(s, d, metadata, symlinks, ignore)
        else:
            shutil.copy2(s, d) if metadata else shutil.copy(s, d)

    src_path = src
    dst_path = dst

    try:
        lst = [p.name for p in src_path.iterdir()]
        if not dst_path.exists():
            dst_path.mkdir(parents=True, exist_ok=True)
            if metadata:
                shutil.copystat(src_path, dst_path)
    except NotADirectoryError:  # egg-link files
        copy_file(src_path.parent, dst_path.parent, src_path.name)
        return

    if ignore:
        excl = ignore(src_path, lst)
        lst = [x for x in lst if x not in excl]

    for item in lst:
        copy_file(src_path, dst_path, item)


def parse_s3_url(url: Optional[str]) -> Tuple[str, str]:
    """
    Parses S3 URL.

    Returns bucket (domain) and file (full path).
    """
    bucket = ""
    path = ""
    if url:
        result = urlparse(url)
        bucket = result.netloc
        path = result.path.strip("/")
    return bucket, path


def human_size(num: float, suffix: str = "B") -> str:
    """
    Convert bytes length to a human-readable version
    """
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:
            return "{0:3.1f}{1!s}{2!s}".format(num, unit, suffix)
        num /= 1024.0
    return "{0:.1f}{1!s}{2!s}".format(num, "Yi", suffix)


def string_to_timestamp(timestring: str) -> int:
    """
    Accepts a str, returns an int timestamp.
    """

    ts = None

    # Uses an extended version of Go's duration string.
    try:
        delta = durationpy.from_str(timestring)
        past = datetime.datetime.now(datetime.timezone.utc) - delta
        ts = calendar.timegm(past.timetuple())
        return ts
    except Exception:
        pass

    if ts:
        return ts
    return 0


##
# `init` related
##


def detect_django_settings() -> List[str]:
    """
    Automatically try to discover Django settings files,
    return them as relative module paths.
    """

    matches = []
    cwd = Path.cwd()
    for settings_file in cwd.rglob("*settings.py"):
        if "site-packages" in str(settings_file):
            continue
        package_path = settings_file.relative_to(cwd)
        package_module = ".".join(package_path.parts).replace(".py", "")
        LOG.info(f"Detected Django settings file: {package_module}")
        matches.append(package_module)
    return matches


def detect_flask_apps() -> list[str]:
    """
    Automatically try to discover Flask apps files,
    return them as relative module paths.
    """

    matches = []
    cwd = Path.cwd()
    for py_file in cwd.rglob("*.py"):
        if "site-packages" in str(py_file):
            continue

        with py_file.open("r", encoding="utf-8") as f:
            lines = f.readlines()
            for line in lines:
                app = None

                # Kind of janky..
                if "= Flask(" in line:
                    app = line.split("= Flask(")[0].strip()
                if "=Flask(" in line:
                    app = line.split("=Flask(")[0].strip()

                if not app:
                    continue

                package_path = py_file.relative_to(cwd)
                package_module = ".".join(package_path.parts).replace(".py", "")
                app_module = f"{package_module}.{app}"
                matches.append(app_module)

    return matches


def get_venv_from_python_version() -> str:
    return "python{}.{}".format(*sys.version_info)


def get_runtime_from_python_version() -> str:
    """ """
    if sys.version_info[0] < 3:
        raise ValueError("Python 2.x is no longer supported.")
    else:
        if sys.version_info[1] <= 7:
            raise ValueError("Python 3.7 and below are no longer supported.")
        elif sys.version_info[1] == 8:
            raise ValueError("Python 3.8 and below are no longer supported.")
        elif sys.version_info[1] == 9:
            return "python3.9"
        elif sys.version_info[1] == 10:
            return "python3.10"
        elif sys.version_info[1] == 11:
            return "python3.11"
        elif sys.version_info[1] == 12:
            return "python3.12"
        elif sys.version_info[1] == 13:
            return "python3.13"
        else:
            raise ValueError(f"Python f{'.'.join(str(v) for v in sys.version_info[:2])} is not yet supported.")


##
# Async Tasks
##


def get_topic_name(lambda_name: str) -> str:
    """Topic name generation"""
    return "%s-zappa-async" % lambda_name


##
# Event sources
##


class BaseEventSource:
    """Base class for event sources"""

    def __init__(self, session: boto3.Session, config: Dict[str, Any]) -> None:
        self.session = session
        self._config = config
        self.arn: str = config.get("arn", "")
        self.enabled: bool = config.get("enabled", True)
        self.batch_size: int = config.get("batch_size", 10)

    def add(self, function_arn: str) -> None:
        raise NotImplementedError

    def remove(self, function_arn: str) -> Union[bool, Dict[str, Any], None]:
        raise NotImplementedError

    def status(self, function_arn: str) -> Optional[Dict[str, Any]]:
        raise NotImplementedError

    def update(self, function_arn: str) -> None:
        raise NotImplementedError


class EventSourceMappingMixin(BaseEventSource):
    """Mixin for event sources that use Lambda event source mappings (SQS, DynamoDB, Kinesis)"""

    def __init__(self, session: boto3.Session, config: Dict[str, Any]) -> None:
        super().__init__(session, config)
        self._lambda = session.client("lambda")

    @property
    def batch_window(self) -> int:
        return self._config.get("batch_window", 1 if self.batch_size > 10 else 0)

    def _get_uuid(self, function_arn: str) -> Optional[str]:
        uuid = None
        response = self._lambda.list_event_source_mappings(
            FunctionName=function_arn,
            EventSourceArn=self.arn,
        )
        LOG.debug(response)
        if len(response["EventSourceMappings"]) > 0:
            uuid = response["EventSourceMappings"][0]["UUID"]
        return uuid

    def add(self, function_arn: str) -> None:
        try:
            kwargs = {
                "FunctionName": function_arn,
                "EventSourceArn": self.arn,
                "BatchSize": self.batch_size,
                "Enabled": self.enabled,
            }
            # Add batch window for SQS
            if hasattr(self, "_supports_batch_window") and self._supports_batch_window:
                kwargs["MaximumBatchingWindowInSeconds"] = self.batch_window

            response = self._lambda.create_event_source_mapping(**kwargs)
            LOG.debug(response)
        except Exception:
            LOG.exception("Unable to add event source")

    def enable(self, function_arn: str) -> None:
        self._config["enabled"] = True
        try:
            response = self._lambda.update_event_source_mapping(
                UUID=self._get_uuid(function_arn),
                Enabled=self.enabled,
            )
            LOG.debug(response)
        except Exception:
            LOG.exception("Unable to enable event source")

    def disable(self, function_arn: str) -> None:
        self._config["enabled"] = False
        try:
            response = self._lambda.update_event_source_mapping(
                UUID=self._get_uuid(function_arn),
                Enabled=self.enabled,
            )
            LOG.debug(response)
        except Exception:
            LOG.exception("Unable to disable event source")

    def update(self, function_arn: str) -> None:
        response = None
        uuid = self._get_uuid(function_arn)
        if uuid:
            try:
                kwargs = {
                    "UUID": uuid,
                    "BatchSize": self.batch_size,
                    "Enabled": self.enabled,
                    "FunctionName": function_arn,
                }
                # Add batch window for SQS
                if hasattr(self, "_supports_batch_window") and self._supports_batch_window:
                    kwargs["MaximumBatchingWindowInSeconds"] = self.batch_window

                response = self._lambda.update_event_source_mapping(**kwargs)
                LOG.debug(response)
            except Exception:
                LOG.exception("Unable to update event source")

    def remove(self, function_arn: str) -> Optional[Dict[str, Any]]:
        response = None
        uuid = self._get_uuid(function_arn)
        if uuid:
            response = self._lambda.delete_event_source_mapping(UUID=uuid)
            LOG.debug(response)
        return response

    def status(self, function_arn: str) -> Optional[Dict[str, Any]]:
        response = None
        LOG.debug("getting status for event source %s", self.arn)
        uuid = self._get_uuid(function_arn)
        if uuid:
            try:
                response = self._lambda.get_event_source_mapping(UUID=uuid)
                LOG.debug(response)
            except botocore.exceptions.ClientError:
                LOG.debug("event source %s does not exist", self.arn)
                response = None
        else:
            LOG.debug("No UUID for event source %s", self.arn)
        return response


class SqsEventSource(EventSourceMappingMixin, BaseEventSource):
    """SQS event source implementation"""

    _supports_batch_window = True


class DynamoDBStreamEventSource(EventSourceMappingMixin, BaseEventSource):
    """DynamoDB Stream event source implementation"""

    _supports_batch_window = False


class KinesisEventSource(EventSourceMappingMixin, BaseEventSource):
    """Kinesis event source implementation"""

    _supports_batch_window = False


class S3EventSource(BaseEventSource):
    """S3 event source implementation"""

    def __init__(self, session: boto3.Session, config: Dict[str, Any]) -> None:
        super().__init__(session, config)
        self._lambda = session.client("lambda")
        self._s3 = session.client("s3")
        bucket_name = self.arn.split(":::")[-1]
        self.bucket_name: str = bucket_name
        self.events: List[str] = config.get("events", ["s3:ObjectCreated:*"])
        self.prefix: str = config.get("prefix", "")
        self.suffix: str = config.get("suffix", "")

    def _make_notification_id(self, function_arn: str) -> str:
        return function_arn.split(":")[-1]

    def add(self, function_arn: str) -> None:
        # Add Lambda permission
        try:
            self._lambda.add_permission(
                FunctionName=function_arn,
                StatementId=f"s3-{self.bucket_name}",
                Action="lambda:InvokeFunction",
                Principal="s3.amazonaws.com",
                SourceArn=self.arn,
            )
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] != "ResourceConflictException":
                LOG.exception("Unable to add Lambda permission for S3")

        # Configure bucket notification
        try:
            # Get existing configuration
            try:
                response = self._s3.get_bucket_notification_configuration(Bucket=self.bucket_name)
                config = response
            except botocore.exceptions.ClientError:
                config = {}

            # Add new configuration
            lambda_configs = config.get("LambdaFunctionConfigurations", [])
            new_config: Dict[str, Any] = {
                "Id": self._make_notification_id(function_arn),
                "LambdaFunctionArn": function_arn,
                "Events": self.events,
            }
            if self.prefix:
                new_config["Filter"] = {"Key": {"FilterRules": [{"Name": "prefix", "Value": self.prefix}]}}
            if self.suffix:
                if "Filter" not in new_config:
                    new_config["Filter"] = {"Key": {"FilterRules": []}}
                new_config["Filter"]["Key"]["FilterRules"].append({"Name": "suffix", "Value": self.suffix})

            lambda_configs.append(new_config)
            config["LambdaFunctionConfigurations"] = lambda_configs

            # Remove ResponseMetadata if present
            config.pop("ResponseMetadata", None)

            self._s3.put_bucket_notification_configuration(Bucket=self.bucket_name, NotificationConfiguration=config)
            LOG.debug("Added S3 event source")
        except Exception:
            LOG.exception("Unable to add S3 event source")

    def remove(self, function_arn: str) -> bool:
        try:
            # Remove Lambda permission
            try:
                self._lambda.remove_permission(FunctionName=function_arn, StatementId=f"s3-{self.bucket_name}")
            except botocore.exceptions.ClientError:
                pass

            # Remove bucket notification
            response = self._s3.get_bucket_notification_configuration(Bucket=self.bucket_name)
            config = response
            lambda_configs = config.get("LambdaFunctionConfigurations", [])
            notification_id = self._make_notification_id(function_arn)
            lambda_configs = [c for c in lambda_configs if c.get("Id") != notification_id]
            config["LambdaFunctionConfigurations"] = lambda_configs
            config.pop("ResponseMetadata", None)
            self._s3.put_bucket_notification_configuration(Bucket=self.bucket_name, NotificationConfiguration=config)
            LOG.debug("Removed S3 event source")
            return True
        except Exception:
            LOG.exception("Unable to remove S3 event source")
            return False

    def status(self, function_arn: str) -> Optional[Dict[str, Any]]:
        try:
            response = self._s3.get_bucket_notification_configuration(Bucket=self.bucket_name)
            lambda_configs = response.get("LambdaFunctionConfigurations", [])
            notification_id = self._make_notification_id(function_arn)
            for config in lambda_configs:
                if config.get("Id") == notification_id:
                    return config
            return None
        except Exception:
            LOG.exception("Unable to get S3 event source status")
            return None

    def update(self, function_arn: str) -> None:
        # For S3, update is remove + add
        self.remove(function_arn)
        self.add(function_arn)


class SNSEventSource(BaseEventSource):
    """SNS event source implementation"""

    def __init__(self, session: boto3.Session, config: Dict[str, Any]) -> None:
        super().__init__(session, config)
        self._lambda = session.client("lambda")
        self._sns = session.client("sns")
        self.filters: Optional[Dict[str, Any]] = config.get("filters")

    def add(self, function_arn: str) -> None:
        # Add Lambda permission
        try:
            self._lambda.add_permission(
                FunctionName=function_arn,
                StatementId=f"sns-{self.arn.split(':')[-1]}",
                Action="lambda:InvokeFunction",
                Principal="sns.amazonaws.com",
                SourceArn=self.arn,
            )
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] != "ResourceConflictException":
                LOG.exception("Unable to add Lambda permission for SNS")

        # Subscribe to topic
        try:
            response = self._sns.subscribe(TopicArn=self.arn, Protocol="lambda", Endpoint=function_arn)
            subscription_arn = response["SubscriptionArn"]
            LOG.debug(response)

            # Add filters if specified
            if self.filters and subscription_arn != "PendingConfirmation":
                self._sns.set_subscription_attributes(
                    SubscriptionArn=subscription_arn,
                    AttributeName="FilterPolicy",
                    AttributeValue=json.dumps(self.filters),
                )
        except Exception:
            LOG.exception("Unable to add SNS event source")

    def remove(self, function_arn: str) -> bool:
        # Check if subscription exists and unsubscribe
        subscription_removed = False
        try:
            response = self._sns.list_subscriptions_by_topic(TopicArn=self.arn)
            for subscription in response["Subscriptions"]:
                if subscription["Endpoint"] == function_arn:
                    self._sns.unsubscribe(SubscriptionArn=subscription["SubscriptionArn"])
                    LOG.debug("Removed SNS subscription")
                    subscription_removed = True
                    break
        except Exception:
            LOG.exception("Unable to remove SNS event source")

        # Only remove Lambda permission if we actually had a subscription
        if subscription_removed:
            try:
                self._lambda.remove_permission(FunctionName=function_arn, StatementId=f"sns-{self.arn.split(':')[-1]}")
            except Exception as e:
                LOG.warning(f"Failed to remove Lambda permission for SNS event source {self.arn}: {e.args}")

        return subscription_removed

    def status(self, function_arn: str) -> Optional[Dict[str, Any]]:
        try:
            response = self._sns.list_subscriptions_by_topic(TopicArn=self.arn)
            for subscription in response["Subscriptions"]:
                if subscription["Endpoint"] == function_arn:
                    return subscription
            return None
        except Exception:
            LOG.exception("Unable to get SNS event source status")
            return None

    def update(self, function_arn: str) -> None:
        # For SNS, update means updating filters if they exist
        if self.filters:
            subscription = self.status(function_arn)
            if subscription:
                try:
                    self._sns.set_subscription_attributes(
                        SubscriptionArn=subscription["SubscriptionArn"],
                        AttributeName="FilterPolicy",
                        AttributeValue=json.dumps(self.filters),
                    )
                except Exception:
                    LOG.exception("Unable to update SNS filters")


class CloudWatchEventSource(BaseEventSource):
    """CloudWatch Events (EventBridge) event source implementation"""

    def __init__(self, session: boto3.Session, config: Dict[str, Any]) -> None:
        super().__init__(session, config)
        self._lambda = session.client("lambda")
        self._events = session.client("events")
        self.rule_name: str = config.get("rule_name", config.get("name", ""))
        self.rule_description: str = config.get("rule_description", config.get("description", ""))
        self.pattern: Optional[Union[str, Dict[str, Any]]] = config.get("event_pattern", config.get("pattern"))
        self.schedule: Optional[str] = config.get("schedule_expression", config.get("schedule"))

    def add(self, function_arn: str) -> None:
        # Create or update rule
        try:
            rule_kwargs: Dict[str, Any] = {
                "Name": self.rule_name,
                "State": "ENABLED" if self.enabled else "DISABLED",
            }
            if self.rule_description:
                rule_kwargs["Description"] = self.rule_description
            if self.pattern:
                rule_kwargs["EventPattern"] = json.dumps(self.pattern) if isinstance(self.pattern, dict) else self.pattern
            if self.schedule:
                rule_kwargs["ScheduleExpression"] = self.schedule

            response = self._events.put_rule(**rule_kwargs)
            rule_arn = response["RuleArn"]
            LOG.debug(response)
        except Exception:
            LOG.exception("Unable to create CloudWatch Events rule")
            return

        # Add Lambda permission
        try:
            self._lambda.add_permission(
                FunctionName=function_arn,
                StatementId=f"events-{self.rule_name}",
                Action="lambda:InvokeFunction",
                Principal="events.amazonaws.com",
                SourceArn=rule_arn,
            )
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] != "ResourceConflictException":
                LOG.exception("Unable to add Lambda permission for CloudWatch Events")

        # Add target
        try:
            self._events.put_targets(Rule=self.rule_name, Targets=[{"Id": "1", "Arn": function_arn}])
            LOG.debug("Added CloudWatch Events target")
        except Exception:
            LOG.exception("Unable to add CloudWatch Events target")

    def remove(self, function_arn: str) -> bool:
        # Remove target
        try:
            self._events.remove_targets(Rule=self.rule_name, Ids=["1"])
        except botocore.exceptions.ClientError:
            pass

        # Remove Lambda permission
        try:
            self._lambda.remove_permission(FunctionName=function_arn, StatementId=f"events-{self.rule_name}")
        except botocore.exceptions.ClientError:
            pass

        # Delete rule
        try:
            self._events.delete_rule(Name=self.rule_name)
            LOG.debug("Removed CloudWatch Events rule")
            return True
        except botocore.exceptions.ClientError:
            pass
        return False

    def status(self, function_arn: str) -> Optional[Dict[str, Any]]:
        try:
            response = self._events.describe_rule(Name=self.rule_name)
            # Check if Lambda is a target
            targets = self._events.list_targets_by_rule(Rule=self.rule_name)
            for target in targets["Targets"]:
                if target["Arn"] == function_arn:
                    return response
            return None
        except botocore.exceptions.ClientError:
            return None

    def update(self, function_arn: str) -> None:
        # Update rule if it exists
        if self.status(function_arn):
            self.add(function_arn)


def get_event_source(
    event_source: Dict[str, Any], lambda_arn: str, target_function: str, boto_session: boto3.Session, dry: bool = False
) -> Tuple[BaseEventSource, str]:
    """
    Given an event_source dictionary item, a session and a lambda_arn,
    create and return the appropriate event source object.
    """
    event_source_map: Dict[str, type[BaseEventSource]] = {
        "dynamodb": DynamoDBStreamEventSource,
        "kinesis": KinesisEventSource,
        "s3": S3EventSource,
        "sns": SNSEventSource,
        "sqs": SqsEventSource,
        "events": CloudWatchEventSource,
    }

    arn = event_source["arn"]
    _, _, svc, _ = arn.split(":", 3)

    event_source_class = event_source_map.get(svc, None)
    if not event_source_class:
        raise ValueError("Unknown event source: {0}".format(arn))

    # Handle S3 special case for function ARN
    if svc == "s3":
        split_arn = lambda_arn.split(":")
        arn_front = ":".join(split_arn[:-1])
        arn_back = split_arn[-1]
        function_arn = ":".join([arn_back, target_function])
        lambda_arn = arn_front
    else:
        function_arn = lambda_arn

    event_source_obj = event_source_class(boto_session, event_source)

    return event_source_obj, function_arn


def add_event_source(
    event_source: Dict[str, Any], lambda_arn: str, target_function: str, boto_session: boto3.Session, dry: bool = False
) -> str:
    """
    Given an event_source dictionary, create the object and add the event source.
    """
    event_source_obj, function_arn = get_event_source(event_source, lambda_arn, target_function, boto_session, dry=False)
    # TODO: Detect changes in config and refine exists algorithm
    if not dry:
        if not event_source_obj.status(function_arn):
            event_source_obj.add(function_arn)
            return "successful" if event_source_obj.status(function_arn) else "failed"
        else:
            return "exists"

    return "dryrun"


def remove_event_source(
    event_source: Dict[str, Any], lambda_arn: str, target_function: str, boto_session: boto3.Session, dry: bool = False
) -> Union[BaseEventSource, bool, Dict[str, Any], None]:
    """
    Given an event_source dictionary, create the object and remove the event source.
    """
    event_source_obj, function_arn = get_event_source(event_source, lambda_arn, target_function, boto_session, dry=False)

    if not dry:
        rule_response = event_source_obj.remove(function_arn)
        return rule_response
    else:
        return event_source_obj


def get_event_source_status(
    event_source: Dict[str, Any], lambda_arn: str, target_function: str, boto_session: boto3.Session, dry: bool = False
) -> Optional[Dict[str, Any]]:
    """
    Given an event_source dictionary, create the object and get the event source status.
    """
    event_source_obj, function_arn = get_event_source(event_source, lambda_arn, target_function, boto_session, dry=False)
    return event_source_obj.status(function_arn)


##
# Analytics / Surveillance / Nagging
##


def check_new_version_available(this_version: str) -> bool:
    """
    Checks if a newer version of Zappa is available.

    Returns True is updateable, else False.

    """
    import requests

    pypi_url = "https://pypi.org/pypi/Zappa/json"
    resp = requests.get(pypi_url, timeout=1.5)
    top_version = resp.json()["info"]["version"]

    return this_version != top_version


class InvalidAwsLambdaName(Exception):
    """Exception: proposed AWS Lambda name is invalid"""

    pass


def validate_name(name: str, maxlen: int = 80) -> str:
    """Validate name for AWS Lambda function.
    name: actual name (without `arn:aws:lambda:...:` prefix and without
        `:$LATEST`, alias or version suffix.
    maxlen: max allowed length for name without prefix and suffix.

    The value 80 was calculated from prefix with longest known region name
    and assuming that no alias or version would be longer than `$LATEST`.

    Based on AWS Lambda spec
    http://docs.aws.amazon.com/lambda/latest/dg/API_CreateFunction.html

    Return: the name
    Raise: InvalidAwsLambdaName, if the name is invalid.
    """
    if not isinstance(name, str):
        msg = "Name must be of type string"
        raise InvalidAwsLambdaName(msg)
    if len(name) > maxlen:
        msg = "Name is longer than {maxlen} characters."
        raise InvalidAwsLambdaName(msg.format(maxlen=maxlen))
    if len(name) == 0:
        msg = "Name must not be empty string."
        raise InvalidAwsLambdaName(msg)
    if not re.match("^[a-zA-Z0-9-_]+$", name):
        msg = "Name can only contain characters from a-z, A-Z, 0-9, _ and -"
        raise InvalidAwsLambdaName(msg)
    return name


def contains_python_files_or_subdirs(folder: str) -> bool:
    """
    Checks (recursively) if the directory contains .py or .pyc files
    """
    folder_path = Path(folder)
    # Check for .py files
    if any(folder_path.rglob("*.py")):
        return True
    # Check for .pyc files
    if any(folder_path.rglob("*.pyc")):
        return True
    return False


def conflicts_with_a_neighbouring_module(directory_path: str) -> bool:
    """
    Checks if a directory lies in the same directory as a .py file with the same name.
    """
    dir_path = Path(directory_path).resolve()
    parent_dir = dir_path.parent
    current_dir_name = dir_path.name
    conflicting_neighbour_filename = current_dir_name + ".py"
    conflicting_file = parent_dir / conflicting_neighbour_filename
    return conflicting_file.exists()


# https://github.com/Miserlou/Zappa/issues/1188
def titlecase_keys(d: Dict[str, Any]) -> Dict[str, Any]:
    """
    Takes a dict with keys of type str and returns a new dict with all keys titlecased.
    """
    return {k.title(): v for k, v in d.items()}


# https://github.com/Miserlou/Zappa/issues/1688
def is_valid_bucket_name(name: str) -> bool:
    """
    Checks if an S3 bucket name is valid according to:
     https://docs.aws.amazon.com/AmazonS3/latest/dev/BucketRestrictions.html#bucketnamingrules
    """
    # Bucket names must be at least 3 and no more than 63 characters long.
    if len(name) < 3 or len(name) > 63:
        return False
    # Bucket names must not contain uppercase characters or underscores.
    if any(x.isupper() for x in name):
        return False
    if "_" in name:
        return False
    # Bucket names must start with a lowercase letter or number.
    if not (name[0].islower() or name[0].isdigit()):
        return False
    # Bucket names must be a series of one or more labels. Adjacent labels are separated by a single period (.).
    for label in name.split("."):
        # Each label must start and end with a lowercase letter or a number.
        if len(label) < 1:
            return False
        if not (label[0].islower() or label[0].isdigit()):
            return False
        if not (label[-1].islower() or label[-1].isdigit()):
            return False
    # Bucket names must not be formatted as an IP address (for example, 192.168.5.4).
    looks_like_IP = True
    for label in name.split("."):
        if not label.isdigit():
            looks_like_IP = False
            break
    if looks_like_IP:
        return False

    return True


def merge_headers(event: Dict[str, Any]) -> Dict[str, str]:
    """
    Merge the values of headers and multiValueHeaders into a single dict.
    Opens up support for multivalue headers via API Gateway and ALB.
    See: https://github.com/Miserlou/Zappa/pull/1756
    """
    headers = event.get("headers") or {}
    multi_headers = (event.get("multiValueHeaders") or {}).copy()
    for h in set(headers.keys()):
        if h not in multi_headers:
            multi_headers[h] = [headers[h]]
    for h in multi_headers.keys():
        multi_headers[h] = ", ".join(multi_headers[h])
    return multi_headers


class ApacheNCSAFormatters:
    """
    NCSA extended/combined Log Format:
    "%h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-agent}i\""
    %h: Remote hostname.
    %l: Remote logname
    %u: Remote user if the request was authenticated. May be bogus if return status (%s) is 401 (unauthorized).
    %t: Time the request was received, in the format [18/Sep/2011:19:18:28 -0400].
        The last number indicates the timezone offset from GMT
    %r: First line of request.
    %>s: Final Status
    %b: Size of response in bytes, excluding HTTP headers.
        In CLF format, i.e. a '-' rather than a 0 when no bytes are sent.
    %{Referer}i:The contents of Referer: header line(s) in the request sent to the server.
    %{User-agent}i: The contents of User-agent: header line(s) in the request sent to the server.

    Refer to:
    https://httpd.apache.org/docs/current/en/mod/mod_log_config.html
    """

    @staticmethod
    def format_log(status_code: int, environ: dict, content_length: int, **kwargs) -> str:
        ip_header = kwargs.get("ip_header", None)
        if ip_header:
            host = environ.get(ip_header, "")
        else:
            host = environ.get("REMOTE_ADDR", "")

        logname = "-"
        user = "-"
        now = datetime.datetime.now(datetime.timezone.utc)
        display_datetime = now.strftime("%d/%b/%Y:%H:%M:%S %z")
        method = environ.get("REQUEST_METHOD", "")
        path_info = environ.get("PATH_INFO", "")
        query_string = ""
        raw_query_string = environ.get("QUERY_STRING", "")
        if raw_query_string:
            query_string = f"?{raw_query_string}"
        server_protocol = environ.get("SERVER_PROTOCOL", "")
        request = f"{method} {path_info}{query_string} {server_protocol}"
        referer = environ.get("HTTP_REFERER", "")
        agent = environ.get("HTTP_USER_AGENT", "")
        log_entry = (
            f'{host} {logname} {user} [{display_datetime}] "{request}" {status_code} {content_length} "{referer}" "{agent}"'
        )
        return log_entry

    @staticmethod
    def format_log_with_response_time(*args, **kwargs) -> str:
        """
        Expect that kwargs includes response time in microseconds, 'rt_us'.
        Mimics Apache-like access HTTP log where the response time data is enabled

        NCSA extended/combined Log Format:
            "%h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-agent}i\" %T/%D"

        %T: The time taken to serve the request, in seconds.
        %D: The time taken to serve the request, in microseconds.
        """
        response_time_microseconds = kwargs.get("rt_us", None)
        log_entry = ApacheNCSAFormatters.format_log(*args, **kwargs)
        if response_time_microseconds:
            response_time_seconds = int(response_time_microseconds / 1_000_000)
            log_entry = f"{log_entry} {response_time_seconds}/{response_time_microseconds}"
        return log_entry


def ApacheNCSAFormatter(with_response_time: bool = True) -> Callable:
    """A factory that returns the wanted formatter"""
    if with_response_time:
        return ApacheNCSAFormatters.format_log_with_response_time
    else:
        return ApacheNCSAFormatters.format_log


def validate_json_serializable(*args: Any, **kwargs: Any) -> None:
    try:
        json.dumps((args, kwargs))
    except (TypeError, OverflowError):
        raise UnserializableJsonError("Arguments to asynchronous.task must be JSON serializable!")
